import pivotr.nodes.nodes as nodes
from getpass import getpass

# Get new node info and return new node object
def add_node(pnodes):
    verify_pid  = False                     # parent ID flag
    verify_addr = False                     # IP addr flag
    node_id     = int                       # Initialize node ID as integer
    node_id     = pnodes[-1].node_id + 1    # Node ID is the highest current node ID + 1

    print('[+] Adding node: n' + str(node_id))

    # Verify the node id exists
    while True:
        parent_id = input('Parent node ID: ')
        
        try:
            for node in pnodes:
                if int(parent_id) == node.node_id:
                    verify_pid = True
                    break
            if verify_pid:
                parent_node = get_node_by_id(int(parent_id), pnodes)
                break
            else:
                print('Unable to locate parent node', parent_id)
        except:
            print('Invalid node id:', parent_id)
            
    # Verify address does NOT already exist
    while True:
        addr = input('Node IP address: ')
        verify_addr = True

        for node in pnodes:
            if addr == node.addr:
                print('Host address ' + str(addr) + ' is current associated with another node.')
                verify_addr = False
                break
        if verify_addr: break

    #username    = input('SSH username: ')
    #passwd      = getpass('SSH password: ')

    return nodes.Node(node_id, parent_node, addr)

# Delete node from nodes list
def delete_node(node, pnodes):

    # Confirm node delete before proceeding
    confirm = 'n'

    try:
        confirm = input('Delete node ' + str(node.node_id) + ' and all attached nodes? (y/N): ')
        if confirm == 'y' or confirm == 'Y':

            # Build list of nodes to delete (select node and all subseqent child nodes)
            def build_child_list(node, delete_list):

                # Iterate through childNodes list if any children exist
                if len(node.childNodes) > 0:
                    for child in node.childNodes:

                        # Run build_child_list recursively on next set of children
                        delete_list = build_child_list(child, delete_list)
                        delete_list.append(child)
                else:

                    # Nodes without children are directly appended to delete list
                    delete_list.append(node)
                    return delete_list

                # After all children are added, add current node and return list
                return delete_list

            # Initialize list, generate list, add primary node to list, remove any repeats, and sort in descending order
            delete_list = []
            delete_list = build_child_list(node, delete_list)
            delete_list.append(node)
            delete_list = list(set(delete_list))
            delete_list.sort(key=lambda x: x.node_id, reverse=True)

            # Get parent node and grandparent node objects
            parent = node.parent_node

            # Delete node from parent's childNodes list
            parent.childNodes.remove(node)

            # Iterate through delete_list and delete all items from pnodes
            for n in delete_list:
                print('[+] Deleting node: n' + str(n.node_id))
                pnodes.remove(n)
    except:
        print('No nodes deleted.')

    # Return nodes for updating
    return pnodes

# Verify if we have access to a node (checks for UN/PW in object only, not via connection)
def verify_node_addr(addr, pnodes):
    for node in pnodes:
        if node.addr == addr:
            return True
    return False

# Return node object based on address
def get_node_by_addr(addr, pnodes):
    for n in pnodes:
        if n.addr == addr:
            return n
    return False

# Return object of node ID or 'False' (which should trigger exception in pivotr.py)
def get_node_by_id(node_id, pnodes):
    for n in pnodes:
        if node_id == n.node_id:
            return n
    return False

def test_nodes(pnodes):
    # Test values (with simplified classes)

    counter = int
    counter = 0
    for node in pnodes:
        print(str(counter) + ':', node.node_id, '- ' + node.addr)
        if (len(node.childNodes) > 0):
            for child in node.childNodes:
                print('   ', child.node_id, '-', child.addr)
        counter += 1

# Map out known nodes
def map_nodes(n):

    # Function for printing all nodes in a nodes childNodes list, aka a nodes "column"
    def print_children(node, pnodes, column, indent_string):
        column += 1

        children = node.childNodes
        
        # Build a string that both indents each node and addes appropriate 'connector' branches when necessary
        # Each iteration through print_children() simply adds to the string what's appropriate for the current childNodes list
        # The string state is stored prior to each subsequent recursive call to print_children() so that the function will print
        # the appopriate string once it is returned to upon exiting recursion
        
        # All first gen nodes (nodes directly connected to our pivotr host) only need an indent, no branches
        if children[0].parent_node.node_id == 0:
            indent_string +=  ('           ')
        else:

            # Get parent node and grandparent node objects
            parent = children[0].parent_node
            grandparent = parent.parent_node
        
            # If the parent is not the LAST node of the grandparent's childNodes list then append a '│'
            if not parent == grandparent.childNodes[-1]:
                indent_string +=  ('│           ')

            # Otherwise just extend the indent by one
            else:
                indent_string +=  ('            ')

        # Store this string so that when recursion ends, we'll use this string for the current function instead of the latest string iteration
        stored_string = indent_string

        # Iterate through all of the current node's children
        for child in node.childNodes:

            # If we're printing the last node in a childNode list, give it a bend instead of branching downward
            if child == node.childNodes[-1]:
                print_owned_node(child, 'bend', indent_string)

            # Otherwise, tee it off so the node below connects
            else:
                print_owned_node(child, 'fork', indent_string)

            if (len(child.childNodes) > 0):
                print_children(child, pnodes, column, indent_string)

                # Restore the indent string to its previous state after exiting recursion
                indent_string = stored_string

        column -= 1

    # Function for printing individual node
    def print_owned_node(obj_node, connection, indent_string):

        # If we have creds for the node, print boxes with solid lines
        if obj_node.node_owned():

            # Every line printed for each box type gets an indent string calculated from print_children()
            print(indent_string + '│  ╭─────────────────╮')

            # If this node is not the last in its childNodes list, print a 'fork' connector
            if (connection == 'fork'):
                print(indent_string + '├──┤ n' + str(obj_node.node_id).ljust(15) + '│')
                print(indent_string + '│  │' + obj_node.addr.center(17) + '│')

                # If this node has children, print a 'connector' branch from the bottom of the box
                if (len(obj_node.childNodes) > 0):
                    print(indent_string + '│  ╰────────┬────────╯')
                else:
                    print(indent_string + '│  ╰─────────────────╯')

            # If this node is the last in its childNodes list, print a 'bend' connector
            elif (connection == 'bend'):
                print(indent_string + '╰──┤ n' + str(obj_node.node_id).ljust(15) + '│')
                print(indent_string + '   │' + obj_node.addr.center(17) + '│')
                if (len(obj_node.childNodes) > 0):
                    print(indent_string + '   ╰────────┬────────╯')
                else:
                    print(indent_string + '   ╰─────────────────╯')

        # No creds for node: print dotted line boxes (don't need bottom branch since they won't have children until we get creds)
        else:
            print(indent_string + '│  ╭ ─ ─ ─ ─ ─ ─ ─ ─ ╮')
            if (connection == 'fork'):
                print(indent_string + '├──┤ n' + str(obj_node.node_id).ljust(15) + '╎')
                print(indent_string + '│  ╎' + obj_node.addr.center(17) + '╎')
                print(indent_string + '│  ╰ ─ ─ ─ ─ ─ ─ ─ ─ ╯')
            elif (connection == 'bend'):
                print(indent_string + '╰──┤ n' + str(obj_node.node_id).ljust(15) + '╎')
                print(indent_string + '   ╎' + obj_node.addr.center(17) + '╎')
                print(indent_string + '   ╰ ─ ─ ─ ─ ─ ─ ─ ─ ╯')

    # Print server node first
    print('╭─────────────────╮')
    print('│' + 'n0 - pivotr'.ljust(17) + '│')
    print('│' + n[0].addr.center(17) + '│')
    print('╰──────────┬──────╯')

    column = 0
    indent_string = ''
    
    try:
        print_children(n[0], n, column, indent_string)
    except:
        print('No further data.')
        print('Try adding a node manually (-a) or performing host discovery (-S)')


import os
import datetime
import pickle
from cryptography.fernet import Fernet

def update_nodes(nodes, filename, key):
	decrypt_file(filename, key)
	put_data(nodes, filename)
	encrypt_file(filename, key)

def get_nodes(filename, key):
	decrypt_file(filename, key)
	nodes = get_data(filename)
	encrypt_file(filename, key)
	return nodes

# Serialize data for preservation (used to preserve our nodes list)
def put_data(obj_list, filename):
	with open(filename, 'wb') as f:
		pickle.dump(obj_list, f)

# Obtain our preserved data (nodes list)
def get_data(filename):
	data = []
	if os.path.exists(filename):
		with open(filename, 'rb') as f:
			if os.path.getsize(filename) > 0:
				data = pickle.load(f)
	return data

# Save a copy of /tmp/nodes.data to ./snapshots
def save_data(filename):
	if os.path.exists(filename):
		now = datetime.datetime.now()
		new_filename = now.strftime("%Y-%m-%d_%H%M%S-nodes.data")
		cp_string = "cp " + filename + " ./snapshots/" + new_filename
		try:
			os.popen(cp_string)
			print('[+] Saved node data to snapshots/' + new_filename)
		except:
			print('[-] An error occured. No file saved.')

# Delete /tmp/nodes.data (or file specified by -f flag)
def clear_data(filename):

	# A confirmation before deleting data is always preferred
	confirm = 'n'
	confirm = input('Clear all node data? (y/N): ')
	if confirm == 'y' or confirm == 'y':
		try:
			os.remove(filename)
			print('[+] Node data deleted.')
		except:
			print('[-] Unable to locate ' + filename)
	else:
		print('[-] Node data not deleted.')

# Generate a random encryption key
def generate_key(filename):
	key = Fernet.generate_key()

	# Store our key in the main pivotr directory
	with open(filename, 'wb') as filekey:
		filekey.write(key)

def get_key(filename):
	with open(filename, 'rb') as filekey:
		key = filekey.read()
	return key

# Encrypt the contents of a file using the provided key
def encrypt_file(filename, key):
	cipher_suite = Fernet(key)  # Initialize a Fernet object with the encryption key
	with open(filename, 'rb') as file:
		plaintext = file.read()  # Read the plaintext data from the file
	encrypted_data = cipher_suite.encrypt(plaintext)  # Encrypt the plaintext data
	with open(filename, 'wb') as encrypted_file:
		encrypted_file.write(encrypted_data)  # Overwrite existing encrypted file with new data

# Decrypt the contents of an encrypted file using the provided key
def decrypt_file(filename, key):
	cipher_suite = Fernet(key)  # Initialize a Fernet object with the key
	with open(filename, 'rb') as encrypted_file:
		encrypted_data = encrypted_file.read()  # Read the encrypted data from the file
	decrypted_data = cipher_suite.decrypt(encrypted_data)  # Decrypt the data
	with open(filename, 'wb') as decrypted_file:
		decrypted_file.write(decrypted_data)  # Write the decrypted data


from scp import SCPClient
import paramiko
import sys
import os
import subprocess
import select
import socket
import logging
import termios
import tty
import pivotr.nodes.node_funcs as node_funcs
import pivotr.networking.networking_funcs as networking_funcs

# Establish a tunnel between pivotr and final host in ip_chain list, returns final host client
def build_ssh_tunnel(ip_chain, pnodes):

    # Establish channels between nodes recursively
    def build_tunnel(c, addr, hosts, counter):

        if (addr == ip_chain[-1]):
            return c, hosts, counter

        # Build tunnel between previous node and next node in ip_chain list
        if (counter == 0):
            prev_node = (pnodes[0].addr, 22)
        else:
            prev_node = (ip_chain[counter - 1], 22)
        next_node = (ip_chain[counter + 1], 22)

        c_transport = c.get_transport()
        c_channel = c_transport.open_channel("direct-tcpip", next_node, prev_node)

        node = node_funcs.get_node_by_addr(ip_chain[counter + 1], pnodes)

        if not node.node_owned():
            node.get_creds()
        try:
            client = paramiko.SSHClient()
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            if node.ssh_keyfile:
                pkey = paramiko.RSAKey.from_private_key_file(node.ssh_keyfile)
                #client.connect(next_node[0], username=node.username, key_filename=node.ssh_keyfile, sock=c_channel)
                client.connect(next_node[0], username=node.username, pkey=pkey, sock=c_channel)
            else:
                client.connect(next_node[0], username=node.username, password=node.passwd, sock=c_channel)

            #client.connect(next_node, username=node.username, password=node.passwd, sock=c_channel)

        except paramiko.ssh_exception.AuthenticationException:
            print('Authentication failed at node n' + str(node.node_id))
            sys.exit(2)

        hosts.append(client)

        counter += 1
        client, hosts, counter = build_tunnel(client, ip_chain[counter], hosts, counter)
        return client, hosts, counter

    # Setup a logger to announce only critical errors with SSH connection (prevents spam from ping_sweep)
    logging.basicConfig()
    logging.getLogger("paramiko").setLevel(logging.CRITICAL)
    counter = 0
    hosts   = []
    node    = node_funcs.get_node_by_addr(ip_chain[counter], pnodes)

    # Build our client object based on creds from node, current chain address and next address
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        if node.ssh_keyfile:
            client.connect(ip_chain[counter], username=node.username, key_filename=node.ssh_keyfile)
        else:
            client.connect(ip_chain[counter], username=node.username, password=node.passwd)
    except paramiko.ssh_exception.AuthenticationException:
        print('Authentication failed at node n' + str(node.node_id))
        sys.exit(2)
    except:
        print('[-] Unable to establish SSH connection')
        sys.exit(2)

    if (len(ip_chain) == 1):
        return client, hosts

    client, hosts, counter = build_tunnel(client, ip_chain[counter], hosts, counter)
    return client, hosts

# Begin SSH server, used for sending commands, interactive sessions, and using SCP for file transfers
def ssh_server(node):
    client = paramiko.SSHClient()

    # client can also support using key files
    # client.load_host_keys('/home/user/.ssh/known_hosts')

    # Establish a connection to target server
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        client.connect(node.addr, username=node.username, password=node.passwd)
    except paramiko.ssh_exception.AuthenticationException:
        print('Authentication failed at node n' + str(node.node_id))
        sys.exit(2)
    except:
        print('[-] Unable to establish SSH connection')
        sys.exit(2)

    return client

def scp_get(node, pnodes, filename):

    ip_chain = networking_funcs.build_ip_chain(node, pnodes)

    client, hosts_to_close = build_ssh_tunnel(ip_chain, pnodes)
    scp = SCPClient(client.get_transport())
    scp.get(r'' + filename, r'./')

    # Close all connections made in our 'ssh chain'
    for session in reversed(hosts_to_close):
        session.close()

def scp_put(node, pnodes, filename):

    ip_chain = networking_funcs.build_ip_chain(node, pnodes)

    client, hosts_to_close = build_ssh_tunnel(ip_chain, pnodes)
    scp = SCPClient(client.get_transport())
    scp.put(filename, recursive=True, remote_path='/tmp')

    # Close all connections made in our 'ssh chain'
    for session in reversed(hosts_to_close):
        session.close()

def ssh_command(node, pnodes, command):

    ip_chain = networking_funcs.build_ip_chain(node, pnodes)
    client, hosts_to_close = build_ssh_tunnel(ip_chain, pnodes)

    stdin, stdout, stderr = client.exec_command(command)
    output = (stdout.read().decode('utf-8'))
    print(output)

    # Close all connections made in our 'ssh chain'
    for session in reversed(hosts_to_close):
        session.close()
    return output

# Establish a live SSH connection to last host in ip_chain list
def ssh_interactive(node, pnodes):

    # Shell function
    def open_shell(connection):

        oldtty_attrs = termios.tcgetattr(sys.stdin)

        # invoke_shell with default options is vt100 compatible
        # which is exactly what you want for an OpenSSH imitation
        channel = connection.invoke_shell()

        def resize_pty():
            # resize to match terminal size
            tty_height, tty_width = \
                    subprocess.check_output(['stty', 'size']).split()

            # try to resize, and catch it if we fail due to a closed connection
            try:
                channel.resize_pty(width=int(tty_width), height=int(tty_height))
            except paramiko.ssh_exception.SSHException:
                pass

        try:
            stdin_fileno = sys.stdin.fileno()
            tty.setraw(stdin_fileno)
            tty.setcbreak(stdin_fileno)

            channel.settimeout(0.0)

            is_alive = True

            while is_alive:
                # resize on every iteration of the main loop
                resize_pty()

                read_ready, write_ready, exception_list = \
                        select.select([channel, sys.stdin], [], [])

                # if the channel is one of the ready objects, print
                # it out 1024 chars at a time
                if channel in read_ready:
                    # try to do a read from the remote end and print to screen
                    try:
                        out = channel.recv(1024).decode('utf-8')

                        # remote close
                        if len(out) == 0:
                            is_alive = False
                        else:
                            # rely on 'print' to correctly handle encoding
                            print(out, end='')
                            sys.stdout.flush()

                    # do nothing on a timeout, as this is an ordinary condition
                    except socket.timeout:
                        pass

                # if stdin is ready for reading
                if sys.stdin in read_ready and is_alive:

                    char = os.read(stdin_fileno, 1)

                    # if this side of the connection closes, shut down gracefully
                    if len(char) == 0:
                        is_alive = False
                    else:
                        channel.send(char)

            channel.shutdown(2)

        finally:
            termios.tcsetattr(sys.stdin, termios.TCSAFLUSH, oldtty_attrs)
            print('[+] SSH connection closed.')

    ip_chain = networking_funcs.build_ip_chain(node, pnodes)

    # Build an SSH tunnel from end to end of the ip_chain
    client, hosts_to_close = build_ssh_tunnel(ip_chain, pnodes)
    
    open_shell(client)
    client.close()

    # Close all connections made in our 'ssh chain'
    for session in reversed(hosts_to_close):
        session.close()



#!/usr/bin/env python
#
##############################################################################################################
# Author: Maturon Miner III
# License: GPLv3
#
# Pivotr - This project aims to establish seemless connectivity between compromised machines on a network
# via SSH tunneling while also providing host discovery tools as well as a CLI-based illustration of all
# known hosts connected, directly or indirectly, to our primary host machine.
#
# Phase 1 - Establish a wholly functional network-based SSH routing infrastructure while supplying tools
# for continued mapping of the network itself and subnets that we contact through compromised nodes. This
# phase relies on credentials being obtained through other means and does NOT actually provide any sort of
# exploitation devices of its own and is functional soley as a convenience for organization of already
# compromised hosts. This phase was more or less completed on 02/01/2024.
#
# Phase 2 - Enact agent executables that, when uploaded and executed on targets, have callback features which
# make reverse-shell connections, maintain persistence, and obfuscate and exfiltrate data to a 'server'
# daemon that is run/modified by the pivotr.py application. Agent features and the data they obtain will
# be integrated into pivotr's current mapping features and data maintenance functionality, and current node
# objects will expand greatly to incorporate the new information available to us through these agents.
#
##############################################################################################################

import os
import sys
import argparse
import pivotr.file.file_funcs as file_funcs
import pivotr.nodes.nodes as nodes
import pivotr.nodes.node_funcs as node_funcs
import pivotr.networking.ssh_funcs as ssh_funcs
import pivotr.networking.networking_funcs as networking_funcs

VERSION = '0.0.2-6'

def usage():
	print('pivotr v'+ VERSION)
	print('Usage: pivotr <ARGS> [OPTIONS]')
	print('')
	print('[ARGS]')
	print('  -n,--node=<NODE_ID>'.ljust(30)  + 'Perform node-related action (no further args prompts to add new node')
	print('  -m,--map'.ljust(30)  + 'Display map of all active nodes')
	print('  -s,--snapshot'.ljust(30)  + 'Save an encrypted snapshot of current node data to snapshots directory')
	print('  -f,--filename'.ljust(30)  + 'Specify a different node data file. Changes take place with next pivotr command')
	print('  -C,--clear'.ljust(30)  + 'Delete all node data')
	print('[NODE OPTIONS]')
	print('  <NO NODE ID>'.ljust(30)  + 'Add a new node')
	print('  <NO FLAG>'.ljust(30)  + 'View all data corresponding to node <NODE_ID>')
	print('  -e,--edit'.ljust(30)  + 'Modify node <NODE_ID>')
	print('  -D,--delete'.ljust(30)  + 'Delete node <NODE_ID>')
	print('  -c,--command=<COMMAND>'.ljust(30)  + 'Run command <COMMAND> from node <NODE_ID>')
	print('  -i,--interactive'.ljust(30)  + 'Open SSH connection on node <NODE_ID>')
	print('  -u,--upload'.ljust(30)  + 'Upload file <FILE> to /tmp directory of node <NODE_ID>')
	print('  -d,--download'.ljust(30)  + 'Download file <FILE> from node <NODE_ID>')
	print('  -S,--scan'.ljust(30)  + 'Ping sweep subnet of node <NODE_ID> and add all discoveries to nodes list')
	#print('  -P,--port-forward'.ljust(30) + 'Port forward <PARENT_NODE_ID>:<PORT> --> <CHILD_NODE_ID>:<PORT>')
	print('[MISC OPTIONS]')
	#print( '  -d,--device=<DEVICE>'.ljust(30)  + 'Specify networking device <DEVICE>')
	print('  -p,--port=<PORT>'.ljust(30)  + 'Specify port <PORT> for any commands that utilize a port')
	print('  -v,--version'.ljust(30)  + 'Print version and exit')
	print('  -h,--help'.ljust(30)  + 'Print help and exit')
	print('')
	print('[EXAMPLES]')
	print('  pivotr -n'.ljust(40) + 'Add a new node')
	print('  pivotr -n9'.ljust(40) + 'View all data corresponding to node #9')
	print('  pivotr -n7 -D'.ljust(40) + 'Delete node #7 and any of its subsequent child nodes')
	print('  pivotr -n1 -c \'ping 10.10.11.120\''.ljust(40) + 'Run \'ping 10.10.11.120\' from node #1')
	print('  pivotr -n2 -i'.ljust(40) + 'Open an interactive SSH session on node #2')
	print('  pivotr -C'.ljust(40) + 'Delete all nodes')
	sys.exit()

def main():

	# Vars
	server_addr, cidr 	= networking_funcs.get_network_info()
	port 				= 6666
	pnodes 				= []
	config_dir			= os.path.expanduser("~") + '/.config/pivotr'
	data_file 			= config_dir + '/nodes.data'
	key_file 			= config_dir + '/pivotr.key'

	# If we don't have a '~/.config/pivotr' directory, make it (in order to store config file and data)
	try:
		os.mkdir(config_dir)
	except:
		pass

	# Generate a key for encrypting/decrypting our data file and store in the root pivotr directory. Set restrictive permissions
	if (not os.path.exists(key_file)):
		print('No key file found. Generating pivotr.key for data encryption/decryption.')
		file_funcs.generate_key(key_file)
		os.chmod(key_file, 0o600)

	# Initialize our key variable for use
	key = file_funcs.get_key(key_file)

	# Setup command line args
	parser = argparse.ArgumentParser(description='Pivotr - A remote command execution framework', add_help=False)
	group1 = parser.add_mutually_exclusive_group()
	group2 = parser.add_mutually_exclusive_group()
	parser.add_argument('--port', '-p', type=int, nargs='*')
	group1.add_argument('--filename', '-f', nargs='*')
	group1.add_argument('--help', '-h', action='store_true')
	group1.add_argument('--node', '-n', nargs='*')
	group1.add_argument('--add', '-a', action='store_true')
	group1.add_argument('--map', '-m', action='store_true')
	group1.add_argument('--clear', '-C', action='store_true')
	group1.add_argument('--snapshot', '-s', action='store_true')
	group1.add_argument('--version', '-v', action='store_true')
	group2.add_argument('--command', '-c', nargs='*')
	group2.add_argument('--interactive', '-i', action='store_true')
	group2.add_argument('--edit', '-e', action='store_true')
	group2.add_argument('--delete', '-D', action='store_true')
	group2.add_argument('--upload', '-u', nargs='*')
	group2.add_argument('--download', '-d', nargs='*')
	#group2.add_argument('--port-forward', '-P', nargs='*')
	group2.add_argument('--scan', '-S', action='store_true')

	parser.add_argument('--test', '-t', action='store_true')

	# Parse command line args
	args = parser.parse_args()

	# If there is no data file or it has no data, create it and add our server. Otherwise, unpickle data and save in pnodes list
	try:
		if (not os.path.exists(data_file) or os.path.getsize(data_file == 0)):
			pnodes.append(nodes.Server_Node(server_addr))
			file_funcs.put_data(pnodes, data_file)
			file_funcs.encrypt_file(data_file, key)
		else:
			pnodes = file_funcs.get_nodes(data_file, key)
	except:
		print('Error creating/editing node data file. Exiting.')
		sys.exit(2)

	# Use port other than default
	if args.port:
		port = args.port[0]

	# Print help and exit
	if args.help:
		usage()

	elif args.version:
		print('pivotr - v' + VERSION)
		sys.exit(0)

	# Copy encrypted node data file to /tmp, decrypt and extract data, then encrypt
	elif args.filename:
		try:
			os.popen('cp ' + args.filename[0] + ' ' + data_file)
			print('Now using ' + args.filename[0] + ' for future pivotr data.')
			print('Effects will take place after this command.')
		except:
			print('Error opening ' + data_file)
			sys.exit(2)

	# Save our current nodes data to snapshots directory
	elif args.snapshot:
		file_funcs.save_data(data_file)

	# Map out all current nodes
	elif args.map:
		node_funcs.map_nodes(pnodes)

	# Clear all node data (confirm before executing)
	elif args.clear:
		file_funcs.clear_data(data_file)

	# Debugging function
	elif args.test:
		node_funcs.test_nodes(pnodes)

	# If args.node is a list with no values, treat as 'add new node'
	elif type(args.node) is list:
		if len(args.node) == 0:
			pnodes.append(node_funcs.add_node(pnodes))
			file_funcs.update_nodes(pnodes, data_file, key)

		# If list is populated, determine list args and treat accordingly
		else:

			# Convert node_id arg to object
			n = node_funcs.get_node_by_id(int(args.node[0]), pnodes)

			# Make sure the node exists before doing anything else
			if not n:
				print('n' + str(args.node[0]) + ': node does not exist.')
				sys.exit(2)

			# List of commands to run against our node object
			if args.interactive:
				ssh_funcs.ssh_interactive(n, pnodes)
			elif args.command:
				ssh_funcs.ssh_command(n, pnodes, args.command[0])
			elif args.edit:
				n.edit_node(pnodes)
				file_funcs.update_nodes(pnodes, data_file, key)
			elif args.delete:
				pnodes = node_funcs.delete_node(n, pnodes)
				file_funcs.update_nodes(pnodes, data_file, key)
			elif args.upload:
				ssh_funcs.scp_put(n, pnodes, args.upload[0])
			elif args.download:
				ssh_funcs.scp_get(n, pnodes, args.download[0])
			elif args.scan:
				networking_funcs.ping_sweep(n, pnodes)
				file_funcs.update_nodes(pnodes, data_file, key)
			else:
				n.view_node()
	else:
		usage()


from getpass import getpass
import os

class Server_Node:
	def __init__(self, addr, node_id=0):
		self.node_id 	= node_id 			# Server node ID, defaults to '0'
		self.addr 		= addr 				# Ip address that the node resides on
		self.childNodes	= []				# List of child nodes that belong to this node

	def view_node(self):
		print('Node ID:   ' + str(self.node_id))
		print('IP addr:   ' + self.addr)

class Node:
	def __init__(self, node_id, parent_node, addr):
		self.node_id 		= node_id
		parent_node.childNodes.append(self)		# Append this node to another node's childNodes list
		self.parent_node 	= parent_node 		# Store parent node
		self.addr 			= addr
		self.username 		= '' 				# Default username from creds{}
		self.passwd 		= '' 				# Corresponding default password from creds{}
		self.ssh_keyfile	= ''				# Private key location (if any)
		self.creds 			= {}				# Stores creds as {'username':['password', 'private_key_location']}
		self.childNodes		= []				# List of child nodes that belong to this node

	# View node attributes
	def view_node(self):
		print('╭─────────────────╮')
		print('│ n' + str(self.node_id).ljust(15) + '│')
		print('│' + self.addr.center(17) + '│')
		print('╰───┬─────────────╯')
		print('    │ Attributes:')
		print('    ╰─┬─Node ID:        ' + str(self.node_id))
		print('      ├─Parent ID:      ' + str(self.parent_node.node_id))
		print('      ├─IP address:     ' + self.addr)
		print('      ├─Child nodes:')

		# Iterate through all child nodes and print them accordingly
		for child in self.childNodes:
			if len(self.childNodes) == 0:
				print('      │    ╰─n/a')
				break
			if child == self.childNodes[-1]:
				print('      │    ╰─n' + str(child.node_id) + ' - [' + child.addr + ']')
			else:
				print('      │    ├─n' + str(child.node_id) + ' - [' + child.addr + ']')

		print('      ╰─Credentials:')
		
		# Iterate through our credentials and print them accordingly
		if not self.creds.keys():
			print('           ╰─n/a')
		else:
			for un, pw in self.creds.items():

				# Append a (default) tag to the credentials currently in use for this node
				if un == self.username:
					display_name = un + ' (default)'
				else:
					display_name = un
				if un == list(self.creds.keys())[-1]:
					print('           ╰─' + display_name)
				else:
					print('           ├─' + display_name)
		#print('DEBUG - creds list: ' + str(self.creds))

	# Modify values of node
	def edit_node(self, pnodes):

		# Print a 'banner' for the node we're working on
		print('╭─────────────────╮')
		print('│ n' + str(self.node_id).ljust(15) + '│')
		print('│' + self.addr.center(17) + '│')
		print('╰───┬─────────────╯')

		while True:
			print('    │ Edit:')
			print('    ╰─┬─(a)ddress')
			print('      ├─(c)redentials')
			print('      ╰─(e)xit')
			print('')
			choice = input("Selection: ")
			match choice:

				# Modify the node's IP address (may make self.addr a list eventually to differentiate multi-homed hosts)
				case 'a' | 'A':
					self.addr = input("New node address: ")

				# Enter our credential-editing cases
				case 'c' | 'C':
					while True:

						# If we don't have credentials, run self.get_creds() to add new ones
						if not self.creds.keys():
							print('No credentials listed for this node.')
							self.get_creds()
							break
						count = 1
						name_to_edit = ''
						print('    │ Credentials:')

						# Iterate through node credentials and display them
						for username, passwd in self.creds.items():
							if username == self.username:
								display_name = username + ' (default)'
							else:
								display_name = username
							if count == 1:
								print('    ╰─┬─(' + str(count) + ') ' + display_name)
							else:
								print('      ├─(' + str(count) + ') ' + display_name)
							count += 1
						print('      ├─(a)dd credentials')
						print('      ╰─(b)ack')
						print('')

						choice = input('Selection: ')
						if choice == 'a' or choice == 'A':
							self.get_creds()
							break
						if choice == 'b' or choice == 'B':
							break
						count = 1

						# Iterate through node credentials again to match user choice to credentials picked
						for username, passwd in self.creds.items():
							if int(choice) == count:
								name_to_edit = username
								#pw_to_edit = passwd
								break
							count += 1

						# Give modification options for credentials
						print('    │ Modify credentials:')
						print('    ╰─┬─(m)ake default')
						print('      ├─(e)dit password')
						print('      ├─(a)dd private key')
						print('      ├─(d)elete')
						print('      ╰─(b)ack')
						print('')

						choice = input('Selection: ')
						match choice:

							# Reassign our username and passwd default values
							case 'm' | 'M':
								self.username = name_to_edit
								self.passwd = self.creds[name_to_edit][0]
								break

							# Get new UN/PW, create the new key, delete the old key, and apply new password to new key
							case 'e' | 'E':
								self.creds[name_to_edit][0] = getpass('New password: ')
								if (name_to_edit == self.username):
									self.passwd = self.creds[name_to_edit][0]
								break

							# Specify private key location (in case of keyfile login)
							case 'a' | 'A':
								self.creds[name_to_edit][1] = input("Private key location: ")
								if (name_to_edit == self.username):
									self.ssh_keyfile = self.creds[name_to_edit][1]
								break

							# Delete the set of credentials entirely
							case 'd' | 'D':
								del self.creds[name_to_edit]
								break
							case 'b' | 'B':
								break

				case 'e' | 'E':
					break
				case _:
					print('Invalid choice')
					continue
		print('[+] Node n' + str(self.node_id) + ' successfully updated.')

	# Store new credentials for a node
	def get_creds(self):
		print('Adding credentials to n' + str(self.node_id) + ' - ' + str(self.addr) + ':')
		new_username = input("New username: ")
		new_passwd = getpass("New password: ")
		self.creds[new_username] = [new_passwd, '']

		# Give option to make these credentials the default for this node
		default = input('Make this the default login for this node? (Y/n): ')
		if (default == 'y' or default == 'Y'):
			self.username = new_username
			self.passwd = new_passwd
			print('Setting to default')
		else:
			print('Default credentials remain unchanged')

	"""
	# Return selected credentials to use for a node
	def use_creds(self):

		# No need to list anything or request input if only one set of credentials exists. Simply return the un/pw
		if len(self.creds.items()) == 1:
			username = list(self.creds.keys())[0]
			passwd = list(self.creds.values())[0]
			return username, passwd

		# If we have multiple login credentials, have user choose which set to use and return the un/pw
		while True:
			count = 1
			print('Credentials for n' + str(self.node_id) + ' - ' + str(self.addr) + ':')
			for username, passwd in self.creds.items():
				print(str(count) + ') ' + username)
				count += 1
			choice = int(input('Number to use: '))
			if choice < 1 or choice > count:
				print('Invalid input')
				continue
			count = 1
			for username, passwd in self.creds.items():
				if choice == count:
					return username, passwd
				count += 1
	"""

	# Return True if we have credentials for the object addr and False otherwise
	def node_owned(self):
		if self.creds.keys():
			for username, passwd in self.creds.items():
				if username:
					if passwd[0] or passwd[1]:
						return True
		return False


import re
import subprocess
import netifaces
import ipaddress
import concurrent.futures
import pivotr.nodes.nodes as nodes
import pivotr.nodes.node_funcs as node_funcs
import pivotr.networking.ssh_funcs as ssh_funcs

# Build chain of hosts from server to node
def build_ip_chain(node, pnodes):
    ip_chain = []

    # Build chain
    def build_list(node, pnodes, ip_chain):

        # Append current node address to beginning of ip_chain list to maintain proper order of execution
        ip_chain.insert(0, node.addr)
        
        # If the parent node of the current node is node_id 0 (server) then return our chained list of addresses
        if node.parent_node.node_id == 0:
            return ip_chain
        else:
            # Otherwise add current node's address and run this function recursively
            build_list(node.parent_node, pnodes, ip_chain)
            return ip_chain

    # Run our recursive building function and return the final list
    ip_chain = build_list(node, pnodes, ip_chain)
    return ip_chain

# Return ip and cidr info
def get_network_info():

    # Get interfaces, then IP from default interface
    ifaces = netifaces.interfaces()
    ip_addr = ''
    subnet_addr = ''
    cidr = 24
    
    # Iterate through devices
    for iface in ifaces:
        try:
            # Pull IP associated with each device, use non-loopback IP
            ip_addr = netifaces.ifaddresses(iface)[netifaces.AF_INET][0]["addr"]
            if ip_addr != '127.0.0.1':

                # Calculate subnet mask and derive CIDR from mask
                subnet_addr = netifaces.ifaddresses(iface)[netifaces.AF_INET][0]["netmask"]
                cidr = sum(bin(int(x)).count('1') for x in subnet_addr.split('.'))
        except:
            pass

    return ip_addr, cidr

# Return all valid IPs of last node in ip_chain list (useful for multi-home hosts)
def get_all_addrs(node, pnodes):

    # Get list of nodes to pivot through
    ip_chain = build_ip_chain(node, pnodes)

    # Establish our SSH tunnel to the last node in the chain
    client, hosts_to_close = ssh_funcs.build_ssh_tunnel(ip_chain, pnodes)

    try:
        # Run 'ip addr' on last host in ip_chain list and capture output
        stdin, stdout, stderr = client.exec_command('ip addr')
        output = (stdout.read().decode('utf-8'))
    except:
        pass

    # Strip all valid IP addresses from output
    ip_addrs = re.findall( r'[0-9]+(?:\.[0-9]+){3}+\/[0-9]{1,2}', output)

    # Iterate through list and remove loopback
    candidates = []
    for addr in ip_addrs:
        candidate = addr.strip('inet ')
        if candidate == '127.0.0.1/8':
            continue
        candidates.append(candidate)

     # Close all connections made in our 'ssh chain'
    for session in reversed(hosts_to_close):
        session.close()

    return candidates
    

# Ping sweep from a given node and add discovered hosts to 
def ping_sweep(node, pnodes):
    
    # If pinging pivotr host's subnet, no need to build an SSH tunnel
    if node.node_id == 0:

        # Simple ping function that appends successful pings to active_hosts list
        def ping(ip_addr):
            try:
                subprocess.check_output(["ping", "-c", "1", ip_addr])
                active_hosts.append(ip_addr)
            except:
                pass

        # Obtain node IP and CIDR info
        ip_addr, cidr = get_network_info()

        # Establish network info based on IP/CIDR
        network = ipaddress.ip_network(ip_addr + '/' + str(cidr), strict=False)
        hosts = network.hosts()
        active_hosts = []

        print('[+] Performing host discovery on n' + str(node.node_id) + ' (' + node.addr + ') subnet')

        # Ping sweep with threading and store live hosts in hosts list
        executor = concurrent.futures.ThreadPoolExecutor(254)
        for candidate_addr in hosts:
            executor.submit(ping, str(candidate_addr))

    else:

        # Ping function, modified to go through our SSH tunnel
        def ping(ip_addr, client):
            try:
                stdin, stdout, stderr = client.exec_command('ping -c1 ' + str(ip_addr))
                if (stdout.read()):
                    active_hosts.append(ip_addr)
            except ssh_exception.NoValidConnectionsError:
                print('No connections')

        # Link all nodes between pivotr host and final host in ip_chain list
        ip_chain = build_ip_chain(node, pnodes)
        client, hosts_to_close = ssh_funcs.build_ssh_tunnel(ip_chain, pnodes)

        # Get all ip addresses associated with node (in case it is multi-homed)
        addrs = get_all_addrs(node, pnodes)
        active_hosts = []

        # Establish our node connection
        client, hosts_to_close = ssh_funcs.build_ssh_tunnel(ip_chain, pnodes)

        # Iterate through all IPs associated with node
        for node_ip in addrs:

            print('[+] Performing host discovery on n' + str(node.node_id) + ' (' + node_ip + ') subnet')

            # Build a list of hosts given the IP/CIDR subnet and initialize our list of active hosts
            ip = ipaddress.ip_network(node_ip, strict=False)
            hosts = ip.hosts()
            
            # If subnet of IP is the same as pivotr subnet, skip
            raw_ip = node_ip.rsplit('/')
            if raw_ip[0] == node.addr:
                continue

            executor = concurrent.futures.ThreadPoolExecutor(254)
            for candidate_addr in hosts:
                executor.submit(ping, str(candidate_addr), client)

        # Close all connections made in our 'ssh chain'
        for session in reversed(hosts_to_close):
            session.close()

    # Iterate live hosts list and build nodes for each one
    if len(active_hosts) > 0:
        for h in active_hosts:
            if not node_funcs.verify_node_addr(h, pnodes):
                node_id = int(pnodes[-1].node_id + 1)
                pnodes.append(nodes.Node(node_id, node, h))
                print('[+] Node added: n'+ str(node_id) + ' - ' + h)
    else:
        print('[-] No new hosts discovered.')

